//
//  Fixed++ - a 16.16 Fixed Point Class for Watcom C++ - v0.01
//

// This code is Copyright (C) 1995 by Nigel Brooke

// See README.DOC for a description of what you can and can't do with it.
// (Don't worry, it's pretty leniant).

//      Nigel Brooke (nbrooke@sfu.ca)

// To avoid including this twice
#ifndef _FIXED_HPP_
#define _FIXED_HPP_

// Make sure we are under 32-bit Watcom
#ifndef __WATCOMC__
#define _FLOAT_
#endif
#ifndef __386__
#define _FLOAT_
#endif


#ifndef _FLOAT_

#ifndef __cplusplus
#error fixed.hpp is for use with C++
#endif

// The chunk of assembly that performs the actual
// multiplication of two 16.16 fixed point values
extern long fixedmul(long a, long b);
#pragma aux fixedmul = \
  ".386"               \
  "imul ebx"           \
  "shld edx,eax,16"    \
  parm [eax] [ebx]     \
  value [edx];


// The chunk of assembly that performs the actual
// division of two 16.16 fixed point values
extern long fixeddiv(long a, long b);
#pragma aux fixeddiv = \
  ".386"               \
  "xor eax, eax"       \
  "shrd eax, edx, 16"  \
  "sar edx, 16"        \
  "idiv ebx"           \
  parm [edx] [ebx]     \
  value [eax];


class fixed {

  // The 'real' value (DUH!)
  int value;

public:

  // Various constructors
  fixed() {};
  fixed( fixed const & a) {value = a.value;}
  fixed( int a) { value = a<<16;}
  fixed( float a) { value = (float)a*65536;}
  fixed( double a) { value = (double)a*65536;}


  // For directly manipulationg the real value of the number.
  // **DO NOT USE** if you intend for the code to compile in
  // floating point mode
  inline int getvalue(void) const { return value; }
  inline void setvalue(int a) { value = a; }

  // The cast operators, one for each built in type
  inline operator unsigned char () {return value >> 16;}
  inline operator char () {return value >> 16;}
  inline operator unsigned short () {return value >> 16;}
  inline operator short () {return value >> 16;}
  inline operator unsigned int () {return value >> 16;}
  inline operator int () {return value >> 16;}
  inline operator unsigned long () {return value >> 16;}
  inline operator long () {return value >> 16;}
  inline operator float () { return float(value) / 65536;}
  inline operator double () { return double(value) / 65536;}

  // All the operators where the fixed point number is the Lvalue (some
  // compiler-speak there :-)
  //
  // In general there are four of each operation , one each for fixed,
  // int, float and double. Thankfully, Watcom will automatically cast
  // any other ordinal types (short, long, char, unsigned int, etc.) to int
  inline fixed &operator= (fixed const& a) {value = a.value; return *this;}
  inline fixed &operator= (int a) { value = long(a)<<16; return *this;}
  inline fixed &operator= (float a) { value = (float)a*65536; return *this;}
  inline fixed &operator= (double a) { value = (double)a*65536; return *this;}

  fixed &operator += ( fixed const & a) {value += a.value; return *this;}
  fixed &operator += ( int a) {value += a<<16; return *this;}
  fixed &operator += ( float a) {value += a*65536; return *this;}
  fixed &operator += ( double a) {value += a*65536; return *this;}

  fixed &operator -= ( fixed const & a) {value -= a.value; return *this;}
  fixed &operator -= ( int a) {value -= a<<16; return *this;}
  fixed &operator -= ( float a) {value -= a*65536; return *this;}
  fixed &operator -= ( double a) {value -= a*65536; return *this;}

  fixed &operator *=(fixed const & a) {value = fixedmul(value,a.value); return *this;}
  fixed &operator *=(int a) {value = fixedmul(value,a<<16); return *this;}
  fixed &operator *=(float a) {value = fixedmul(value,a*65536); return *this;}
  fixed &operator *=(double a) {value = fixedmul(value,a*65536); return *this;}

  fixed &operator /=(fixed const & a) {value = fixeddiv(value,a.value); return *this;}
  fixed &operator /=(int a) {value = fixeddiv(value,a<<16); return *this;}
  fixed &operator /=(float a) {value = fixeddiv(value,a*65536); return *this;}
  fixed &operator /=(double a) {value = fixeddiv(value,a*65536); return *this;}

  // The negation operator
  fixed operator - () const { fixed temp; temp.value = -value; return temp;}

};

//  All the binary operators
//
// There are generally seven of each of these, one for each of the four
// types, and then three more for the built in types that simply have
// the arguments reversed

inline fixed operator + (const fixed& a, const fixed& b) {
      fixed temp;
      temp.setvalue(a.getvalue() + b.getvalue());
      return temp;
}

inline fixed operator + ( fixed const & a, int b) {
      fixed temp;
      temp.setvalue(a.getvalue() + b<<16);
      return temp;
}

inline fixed operator + ( fixed const & a, float b) {
      fixed temp;
      temp.setvalue(a.getvalue() + (b*65536));
      return temp;
}

inline fixed operator + ( fixed const &a, double b ){
      fixed temp;
      temp.setvalue(a.getvalue() + (b*65536));
      return temp;
}

inline fixed operator + ( int a, fixed const &b ) {return b+a;}
inline fixed operator + ( float a, fixed const &b ) {return b+a;}
inline fixed operator + ( double a, fixed const &b ) {return b+a;}


inline fixed operator - (const fixed& a, const fixed& b) {
      fixed temp;
      temp.setvalue(a.getvalue() - b.getvalue());
      return temp;
}

inline fixed operator - ( fixed const & a, int b) {
      fixed temp;
      temp.setvalue(a.getvalue() - b<<16);
      return temp;
}

inline fixed operator - ( fixed const & a, float b) {
      fixed temp;
      temp.setvalue(a.getvalue() - (b*65536));
      return temp;
}

inline fixed operator - ( fixed const &a, double b ){
      fixed temp;
      temp.setvalue(a.getvalue() - (b*65536));
      return temp;
}

inline fixed operator - ( int a, fixed const &b ) {return b-a;}
inline fixed operator - ( float a, fixed const &b ) {return b-a;}
inline fixed operator - ( double a, fixed const &b ) {return b-a;}


inline fixed operator *(const fixed& a, const fixed& b) {
      fixed temp;
      temp.setvalue(fixedmul(a.getvalue(),b.getvalue()));
      return temp;
}

inline fixed operator * ( fixed const &a, int b) {
      fixed temp;
      temp.setvalue(fixedmul(a.getvalue(),b<<16));
      return temp;
}

inline fixed operator * ( fixed const &a, float b) {
      fixed temp;
      temp.setvalue(fixedmul(a.getvalue(),b*65536));
      return temp;
}


inline fixed operator * ( fixed const &a, double b) {
      fixed temp;
      temp.setvalue(fixedmul(a.getvalue(),b*65536));
      return temp;
}

inline fixed operator * ( int a, fixed const &b ) {return b*a;}
inline fixed operator * ( float a, fixed const &b ) {return b*a;}
inline fixed operator * ( double a, fixed const &b ) {return b*a;}

inline fixed operator /(const fixed& a, const fixed& b) {
      fixed temp;
      temp.setvalue(fixeddiv(a.getvalue(),b.getvalue()));
      return temp;
}

inline fixed operator / ( fixed const &a, int b) {
      fixed temp;
      temp.setvalue(fixeddiv(a.getvalue(),b<<16));
      return temp;
}

inline fixed operator / ( fixed const &a, float b) {
      fixed temp;
      temp.setvalue(fixeddiv(a.getvalue(),b*65536));
      return temp;
}


inline fixed operator / ( fixed const &a, double b) {
      fixed temp;
      temp.setvalue(fixeddiv(a.getvalue(),b*65536));
      return temp;
}

inline fixed operator / ( int a, fixed const &b ) {return b/a;}
inline fixed operator / ( float a, fixed const &b ) {return b/a;}
inline fixed operator / ( double a, fixed const &b ) {return b/a;}



// The comparison operators
//
// Remember, these cast to fixed, then compare, so. for example
//   fixed(200.1) != int(200.1)
//   (not too surprising, really)
//
// Also fixed point looses acuracy a lot faster than floating point,
// so :
//   fixed(1.0000001) != float(1.0000001)
//   fixed(1.0000001) == int(1.0000001)
//   (a bit more surprising)

inline int operator == ( fixed const &a, fixed const &b ) {
  return a.getvalue() == b.getvalue();
}

inline int operator == ( fixed const &a, int b) {
  return a.getvalue() == b<<16;
}


inline int operator == ( fixed const &a, float b) {
  return a.getvalue() == (int)b*65536;
}

inline int operator == ( fixed const &a, double b) {
  return a.getvalue() == (int)b*65536;
}

inline int operator == ( int a, fixed const &b ) {return b==a;}
inline int operator == ( float a, fixed const &b ) {return b==a;}
inline int operator == ( double a, fixed const &b ) {return b==a;}


inline int operator != ( fixed const &a, fixed const &b ) {
  return a.getvalue() != b.getvalue();
}

inline int operator != ( fixed const &a, int b) {
  return a.getvalue() != b<<16;
}


inline int operator != ( fixed const &a, float b) {
  return a.getvalue() != (int)b*65536;
}

inline int operator != ( fixed const &a, double b) {
  return a.getvalue() != (int)b*65536;
}

inline int operator != ( int a, fixed const &b ) {return b!=a;}
inline int operator != ( float a, fixed const &b ) {return b!=a;}
inline int operator != ( double a, fixed const &b ) {return b!=a;}


inline int operator > ( fixed const &a, fixed const &b ) {
  return a.getvalue() > b.getvalue();
}

inline int operator > ( fixed const &a, int b) {
  return a.getvalue() > b<<16;
}


inline int operator > ( fixed const &a, float b) {
  return a.getvalue() > (int)b*65536;
}

inline int operator > ( fixed const &a, double b) {
  return a.getvalue() > (int)b*65536;
}

inline int operator > ( int a, fixed const &b ) {return b>a;}
inline int operator > ( float a, fixed const &b ) {return b>a;}
inline int operator > ( double a, fixed const &b ) {return b>a;}


inline int operator < ( fixed const &a, fixed const &b ) {
  return a.getvalue() < b.getvalue();
}

inline int operator < ( fixed const &a, int b) {
  return a.getvalue() < b<<16;
}


inline int operator < ( fixed const &a, float b) {
  return a.getvalue() < (int)b*65536;
}

inline int operator < ( fixed const &a, double b) {
  return a.getvalue() < (int)b*65536;
}

inline int operator < ( int a, fixed const &b ) {return b<a;}
inline int operator < ( float a, fixed const &b ) {return b<a;}
inline int operator < ( double a, fixed const &b ) {return b<a;}

inline int operator >= ( fixed const &a, fixed const &b ) {
  return a.getvalue() >= b.getvalue();
}

inline int operator >= ( fixed const &a, int b) {
  return a.getvalue() >= b<<16;
}


inline int operator >= ( fixed const &a, float b) {
  return a.getvalue() >= (int)b*65536;
}

inline int operator >= ( fixed const &a, double b) {
  return a.getvalue() >= (int)b*65536;
}

inline int operator >= ( int a, fixed const &b ) {return b>=a;}
inline int operator >= ( float a, fixed const &b ) {return b>=a;}
inline int operator >= ( double a, fixed const &b ) {return b>=a;}


inline int operator <= ( fixed const &a, fixed const &b ) {
  return a.getvalue() <= b.getvalue();
}

inline int operator <= ( fixed const &a, int b) {
  return a.getvalue() <= b<<16;
}


inline int operator <= ( fixed const &a, float b) {
  return a.getvalue() <= (int)b*65536;
}

inline int operator <= ( fixed const &a, double b) {
  return a.getvalue() <= (int)b*65536;
}

inline int operator <= ( int a, fixed const &b ) {return b<=a;}
inline int operator <= ( float a, fixed const &b ) {return b<=a;}
inline int operator <= ( double a, fixed const &b ) {return b<=a;}

#else
// If _FLOAT_ was defined, or __WATCOMC__ or __386__ weren't
typedef double fixed;

#endif
#endif

